/*  BotikKey Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include <process.h>
#include "opt_tab_mngr_dlg.h"
#include "opt_tab_pwd_dlg.h"
#include "opt_tab_access_dlg.h"
#include "opt_tab_svc_dlg.h"
#include "interface_h.h"
#include "logviewer.h"
#include "dialogs.h"
#include "message_dispatcher.h"
#include "main_dlg.h"

HWND g_hCurTabDlg;
HWND g_hMainDlg;
BOOL g_fExit;

// Forward
//VOID	CALLBACK CheckForNewVersionProc(HWND hwnd_, UINT nMsg_, UINT_PTR idEvent_, DWORD dwTime_);
unsigned int __stdcall CheckForNewVersionProc(void*);

// Processes messages for the main window.
LRESULT CALLBACK WndProcMD(HWND hWnd_, UINT uMsg_, WPARAM wParam_, LPARAM lParam_)
{
	//PAINTSTRUCT ps;
	//HDC hdc;

	static UINT WM_TASKBARCREATED;

	switch (uMsg_) 
	{
	case WM_INITDIALOG:
		WM_TASKBARCREATED = RegisterWindowMessage(_T("TaskbarCreated"));
		RegisterLogViewerWndCls();
		return OnInitDialogMD(hWnd_, wParam_, lParam_);

	case WM_COMMAND:
		OnCommandMD(hWnd_, wParam_, lParam_);
		break;

	case WM_NOTIFY:
		OnNotifyMD(hWnd_, wParam_, lParam_);
		break;

	case WM_TIMER:
		{
			unsigned int thrdid;
			_beginthreadex(NULL, 0, CheckForNewVersionProc, NULL, 0, &thrdid);
		};
		break;

	case WM_SIZE:
		OnSizeMD(hWnd_, wParam_, lParam_);
		break;

	case WM_PAINT:
		//hdc = BeginPaint(hWnd_, &ps);
		//EndPaint(hWnd_, &ps);
		break;

	//case WM_USER_NEW_EVENT:
	//	if(g_fForceLogviewer)
	//		ShowLogViewerWindow();

	//	cfdprintf(_T("WM_USER_NEW_EVENT"));
	//	break;

	case WM_USER_NOTIFYICON:
		OnUserNotifyIconMD(hWnd_, wParam_, lParam_);
		break;

	case WM_USER_KILL_MANAGER:
		cfdprintf("Got WM_USER_KILL_MANAGER");
		DestroyWindow(hWnd_);
		break;

	case WM_USER_SVC_CONNECTED:
		cfdprintf("WM_USER_SVC_CONNECTED");

		if(!lParam_) // If true, skip break to prompt for password.
			break;

	case WM_USER_PROMPT_PASSWORD:
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_PWD_DLG), hWnd_, (DLGPROC)SetPasswdDlg);
		break;

	case WM_USER_SVC_STOPPED:
		cfdprintf("WM_USER_SVC_STOPPED");

		g_nCurAccess = ABK_ACCESS_UNSPEC;		// Set current access variable to unspecified
		DisplayCurrentAccessMode(g_hAccTabDlg);	// Refresh GUI

		g_dwSvcStatus = ABK_SVC_STOPPED;
		DisplayCurrentServiceStatus(g_hAccTabDlg);

		StartMessageDispatching();				// Start MessageDispatcher to connect to service again.
		break;

	case WM_SYSCOMMAND:
		switch(wParam_)
		{
		case SC_MINIMIZE:
			if(g_fHideMin)
			{
				ShowWindow(hWnd_, SW_HIDE);
				return(TRUE);
			};
			break;

		case SC_CLOSE:
			if(!g_fHideClose)
			{
				DoExitManager();
				return(TRUE);
			};
			break;
		};
		break;

	case WM_DESTROY:
		KillTimer(hWnd_, BKP_UPDATE_TIMER_ID);
		DestroyWindow(g_hwndLogViewer);
		DelNotifyIcon(hWnd_, ABK_MNGR_NOTIFY_ICON_ID);
		PostQuitMessage(0);
		break;

	default:
		if(uMsg_ == WM_TASKBARCREATED) // to restore notification icon after explorer.exe crashed.
		{
			AddNotifyIcon(hWnd_, ABK_MNGR_NOTIFY_ICON_ID, IDI_MANAGER);
			DisplayCurrentAccessMode(g_hAccTabDlg);
		};
		break;
	};

	return FALSE;
};

LRESULT OnSizeMD(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	AdjustTabDlg(hWnd_, g_hMngrTabDlg);
	AdjustTabDlg(hWnd_, g_hPwdTabDlg);
	AdjustTabDlg(hWnd_, g_hAccTabDlg);
	AdjustTabDlg(hWnd_, g_hSvcTabDlg);

	return 0;
};

LRESULT OnNotifyMD(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	if(wParam_ == IDC_OPT_DLG_TAB)
		MsgFromTab(hWnd_, wParam_, lParam_);

	return 0;
};

LRESULT OnCommandMD(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	INT wmId    = LOWORD(wParam_);
	INT wmEvent = HIWORD(wParam_);

	// Parse the menu selections:
	switch (wmId)
	{
	case IDM_ABOUT:
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_ABOUT_DLG), hWnd_, (DLGPROC)AboutDlg);
		break;

	case IDM_SHOW_LICENSE:
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_LICENSE_DLG), hWnd_, (DLGPROC)LicenseDlg);
		break;

	case IDM_SET_PASSWD:
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_PWD_DLG), hWnd_, (DLGPROC)SetPasswdDlg);
		break;

	case IDM_SHOW_LOGVIEWER:
		ShowLogViewerWindow();
		break;

	case IDM_SHOW_PWD2KEY:
		{
			RunPwd2KeyUtil();
		};
		break;

	case IDM_OPTIONS:
		ShowWindow(hWnd_, SW_SHOW);
		ShowWindow(hWnd_, SW_RESTORE);
		SetForegroundWindow(hWnd_);
		break;

	case IDM_NET_NONE:
		cfdprintf("Menu: SetAccessMode");
		SetAccessMode(ABK_ACCESS_NONE);

		// Update Access Tab view to display current access mode.
		//DisplayCurrentAccessMode(g_hAccTabDlg);
		return(TRUE);

	case IDM_NET_LOCAL:
		cfdprintf("Menu: SetAccessMode");
		SetAccessMode(ABK_ACCESS_LOCAL);

		// Update Access Tab view to display current access mode.
		//DisplayCurrentAccessMode(g_hAccTabDlg);
		return(TRUE);

	case IDM_NET_WORLD:
		cfdprintf("Menu: SetAccessMode");
		SetAccessMode(ABK_ACCESS_WORLD);

		// Update Access Tab view to display current access mode.
		//DisplayCurrentAccessMode(g_hAccTabDlg);
		return(TRUE);

	case IDM_WEB_SITE:
		ShellExecute(HWND_DESKTOP, _T("open"), ABK_DEV_LINK_SITE, NULL, NULL, SW_SHOWNORMAL);
		return(TRUE);

	case IDHELP:
		{
			TCITEM tci;
			INT nCur;
			HWND hwndTab = GetDlgItem(hWnd_, IDC_OPT_DLG_TAB);
			LPTSTR szPath;
			LPTSTR szTopic;

			tci.mask = TCIF_PARAM;

			nCur = TabCtrl_GetCurSel(hwndTab);
			::SendMessage(hwndTab, TCM_GETITEM, nCur, (LPARAM)&tci);

			if(tci.lParam == IDS_TAB_0002)
				szTopic = _T("::/html/02_tab_01.html");
			else if(tci.lParam == IDS_TAB_0003)
				szTopic = _T("::/html/02_tab_03.html");
			else if(tci.lParam == IDS_TAB_0004)
				szTopic = _T("::/html/02_tab_02.html");
			else
				szTopic = _T("::/html/02_tab_00.html");

			szPath = new TCHAR[_tcslen(g_szChmPath) + _tcslen(szTopic) + 1];

			_tcscpy(szPath, g_szChmPath);
			_tcscat(szPath, _T(szTopic));

			HtmlHelp(HWND_DESKTOP/*g_hMainDlg*/, szPath, HH_DISPLAY_TOPIC, NULL);

			delete[] szPath;
		};
		break;

	case IDM_HELP:
		HtmlHelp(HWND_DESKTOP/*g_hMainDlg*/, g_szChmPath, HH_DISPLAY_TOPIC, NULL);
		break;

	case IDM_EXIT_SHUTDOWN:
		g_fExit = TRUE;

		if(!DoShutdownService())
			DestroyWindow(hWnd_);

		return(TRUE);

	case IDM_EXIT:
		DoExitManager();
		return(TRUE);

	case IDAPPLAY:
		if(!SaveSettings())
			return(TRUE);

		UpdateGUIST(g_hSvcTabDlg);
		UpdateGUIPT(g_hPwdTabDlg);
		UpdateGUIMT(g_hMngrTabDlg);

		g_fSvcConfChanged = FALSE;
		cfEnableDlgItem(g_hMainDlg, IDAPPLAY, FALSE);

		return(TRUE);

	case IDOK:
		if(!SaveSettings())
			return(TRUE);

	case IDCANCEL:
		ShowWindow(hWnd_, SW_MINIMIZE);
		ShowWindow(hWnd_, SW_HIDE);

		UpdateGUIST(g_hSvcTabDlg);
		UpdateGUIPT(g_hPwdTabDlg);
		UpdateGUIMT(g_hMngrTabDlg);

		g_fSvcConfChanged = FALSE;
		cfEnableDlgItem(g_hMainDlg, IDAPPLAY, FALSE);

		return(TRUE);
	};

	return(FALSE);
};

LRESULT OnUserNotifyIconMD(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	DWORD dwMicePos;
	POINT pt;
	HMENU hMenu, hSubMenu, hSubMenu1;
	MENUITEMINFO mii;

	dwMicePos = GetCursorPos(&pt);

	switch(lParam_)
	{
	case WM_LBUTTONUP:
		if(!g_fOneClick)
			break;
	case WM_LBUTTONDBLCLK:
		ShowWindow(hWnd_, SW_SHOW);
		ShowWindow(hWnd_, SW_RESTORE);
		SetForegroundWindow(hWnd_);
		break;

	case WM_MBUTTONUP:
		DialogBox(g_hInst, MAKEINTRESOURCE(IDD_PWD_DLG), hWnd_, (DLGPROC)SetPasswdDlg);
		break;

	case WM_RBUTTONUP:
		cfdprintf("NotifyIcon got WM_RBUTTONDOWN");

		hMenu = LoadMenu(g_hInst, MAKEINTRESOURCE(IDC_MANAGER));
		hSubMenu = GetSubMenu(hMenu, 0);

		mii.cbSize = sizeof(MENUITEMINFO);
		mii.fMask = MIIM_STATE;
		mii.fState = MFS_DEFAULT;

		SetMenuItemInfo(hSubMenu, 0, TRUE, &mii);

		// Change radio items.
		hSubMenu1 = GetSubMenu(hSubMenu, 5);
		if(g_nCurAccess == ABK_ACCESS_UNSPEC)
		{
			mii.fState = MFS_DISABLED;

			SetMenuItemInfo(hSubMenu1, 0, TRUE, &mii);
			SetMenuItemInfo(hSubMenu1, 1, TRUE, &mii);
			SetMenuItemInfo(hSubMenu1, 2, TRUE, &mii);
		}
		else
		{
			DWORD dwPos;

			if(g_nCurAccess & ABK_ACCESS_CHANGING)
				mii.fState = MFS_DISABLED;
			else
				mii.fState = MFS_ENABLED;

			SetMenuItemInfo(hSubMenu1, 0, TRUE, &mii);
			SetMenuItemInfo(hSubMenu1, 1, TRUE, &mii);
			SetMenuItemInfo(hSubMenu1, 2, TRUE, &mii);

			if(g_nCurAccess & ABK_ACCESS_NONE)
				dwPos = 0;
			else if(g_nCurAccess & ABK_ACCESS_LOCAL)
				dwPos = 1;
			else if(g_nCurAccess & ABK_ACCESS_WORLD)
				dwPos = 2;

			CheckMenuRadioItem(hSubMenu1, 0, 2, dwPos, MF_BYPOSITION);
		};

		// NB. We _must_ set our window to foreground before call
		// TrackPopupMenu(Ex), otherwise, the menu will not disappear
		// when user clicks outside of the menu.
		SetForegroundWindow(hWnd_);

		TrackPopupMenuEx(hSubMenu, TPM_LEFTALIGN|TPM_LEFTBUTTON|TPM_VERTICAL, pt.x, pt.y, hWnd_, NULL);

		DestroyMenu(hMenu);
		break;
	};

	return FALSE;
};

LRESULT OnInitDialogMD(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	HWND hWnd;
	TCITEM tci;
	TCHAR szBuf[MAX_LOADSTRING];
	HICON hIcon = 0;

	DWORD dwThrdId = GetWindowThreadProcessId(hWnd_, NULL);
	cfdprintf(_T("Thread Id: %d"), dwThrdId);

	g_hBigIcon = LoadIcon(g_hInst, MAKEINTRESOURCE(IDI_MANAGER));
	SendMessage(hWnd_, WM_SETICON, ICON_BIG, (LPARAM)g_hBigIcon);

	g_hSmallIcon = (HICON)LoadImage(g_hInst, MAKEINTRESOURCE(IDI_MANAGER), IMAGE_ICON, 16, 16, 0);
	SendMessage(hWnd_, WM_SETICON, ICON_SMALL, (LPARAM)g_hSmallIcon);

	//hIcon = (HICON)LoadImage(g_hInst, MAKEINTRESOURCE(IDI_HELP), IMAGE_ICON, 16, 16, 0);
	//SendDlgItemMessage(hWnd_, IDHELP, BM_SETIMAGE, IMAGE_ICON, (LPARAM)hIcon);

	// Create tabs dialogs.
	if(!CreateManagerTab(hWnd_))
		return(FALSE);

	if(!CreatePwdTab(hWnd_))
		return(FALSE);

	if(!CreateAccTab(hWnd_))
		return(FALSE);

	if(!CreateSvcTab(hWnd_))
		return(FALSE);

	if(!CreateLogViewerWnd(hWnd_))
		return(FALSE);

	// Add tabs to the tab control window.
	hWnd = GetDlgItem(hWnd_, IDC_OPT_DLG_TAB);

	tci.mask = TCIF_TEXT|TCIF_PARAM;

	LoadString(g_hInst, IDS_TAB_0001, szBuf, MAX_LOADSTRING-1);
	tci.lParam = IDS_TAB_0001;
	tci.pszText = szBuf;
	tci.cchTextMax = (INT)_tcslen(szBuf);
	SendMessage(hWnd, TCM_INSERTITEM, 0, (LPARAM)&tci);

	LoadString(g_hInst, IDS_TAB_0002, szBuf, MAX_LOADSTRING-1);
	tci.lParam = IDS_TAB_0002;
	tci.pszText = szBuf;
	tci.cchTextMax = (INT)_tcslen(szBuf);
	SendMessage(hWnd, TCM_INSERTITEM, 1, (LPARAM)&tci);

	LoadString(g_hInst, IDS_TAB_0004, szBuf, MAX_LOADSTRING-1);
	tci.lParam = IDS_TAB_0004;
	tci.pszText = szBuf;
	tci.cchTextMax = (INT)_tcslen(szBuf);
	SendMessage(hWnd, TCM_INSERTITEM, 2, (LPARAM)&tci);

	LoadString(g_hInst, IDS_TAB_0003, szBuf, MAX_LOADSTRING-1);
	tci.lParam = IDS_TAB_0003;
	tci.pszText = szBuf;
	tci.cchTextMax = (INT)_tcslen(szBuf);
	SendMessage(hWnd, TCM_INSERTITEM, 3, (LPARAM)&tci);

	SendMessage(hWnd, TCM_SETCURSEL, 0, 0); // Select General tab.
	g_hCurTabDlg = g_hAccTabDlg;
	ShowWindow(g_hCurTabDlg, SW_SHOW);

	AddNotifyIcon(hWnd_, ABK_MNGR_NOTIFY_ICON_ID, IDI_MANAGER);

	g_fSvcConfChanged = FALSE;
	cfEnableDlgItem(hWnd_, IDAPPLAY, FALSE);

	// Make first check for a new version after one second.
	SetTimer(hWnd_, BKP_UPDATE_TIMER_ID, 5000, (TIMERPROC)NULL);
	
	return(StartMessageDispatching());
};

VOID MsgFromTab(HWND hWnd_, WPARAM wParam_, LPARAM lParam_)
{
	LPNMHDR pnmhdr = (LPNMHDR)lParam_;
	INT nCur;
	TCITEM tci;

	switch(pnmhdr->code)
	{
	case TCN_SELCHANGE:
		tci.mask = TCIF_PARAM;
		nCur = TabCtrl_GetCurSel(pnmhdr->hwndFrom);
		::SendMessage(pnmhdr->hwndFrom, TCM_GETITEM, nCur, (LPARAM)&tci);

		ShowWindow(g_hCurTabDlg, SW_HIDE);
		
		if(tci.lParam == IDS_TAB_0002)
			g_hCurTabDlg = g_hMngrTabDlg;
		else if(tci.lParam == IDS_TAB_0003)
			g_hCurTabDlg = g_hPwdTabDlg;
		else if(tci.lParam == IDS_TAB_0004)
			g_hCurTabDlg = g_hSvcTabDlg;
		else
			g_hCurTabDlg = g_hAccTabDlg;

		ShowWindow(g_hCurTabDlg, SW_SHOW);
		break;
	};
};

VOID AdjustTabDlg(HWND hWnd_, HWND hWndDlg_)
{
	RECT rc;
	HWND hWnd;
	
	hWnd = GetDlgItem(hWnd_, IDC_OPT_DLG_TAB);

	// XXX Strange behavior of the tab control...
	// Tab doesn't return right values of GetClientRect.
	// It returns a client rect of parent dialog window
	// instead of self. So, I use a different way to get values.
	GetWindowRect(hWnd, &rc);

	ScreenToClient(hWndDlg_, (LPPOINT)&rc);
	ScreenToClient(hWndDlg_, (LPPOINT)&rc.right);

	//GetClientRect(hWnd, &rc);

	SendMessage(hWnd, TCM_ADJUSTRECT, FALSE, (LPARAM)&rc);

	rc.left -= 2;

	MoveWindow(hWndDlg_, rc.left, rc.top, rc.right-rc.left, 
			   rc.bottom-rc.top, TRUE);
};

BOOL SaveSettings()
{
	BOOL fStartup;
	TCHAR szPath[MAX_PATH+4];
	LONG lLen;
	LPTSTR szBuf;
	DWORD dwVal;
	INT nCurSel;

	// Validate new values.
	dwVal = GetDlgItemInt(g_hSvcTabDlg, IDC_BKEYSRV_PORT, NULL, FALSE);
	if(dwVal > 65535 || dwVal == 0) // Port can't be greater than 65535.
	{
		MessageBox(NULL, _T("  ."), _T("BotikKey Pro..."), MB_OK|MB_ICONHAND);
		SetFocus(GetDlgItem(g_hSvcTabDlg, IDC_BKEYSRV_PORT));
		return(FALSE);
	};

	// Update manager variables.
	// Manager tab.
	if(SendDlgItemMessage(g_hMngrTabDlg, IDC_STARTUP_RADIO1, BM_GETCHECK, 0, 0))
		g_nStartupMode = ABK_MNGR_STARTUP_NORMAL;
	else if(SendDlgItemMessage(g_hMngrTabDlg, IDC_STARTUP_RADIO2, BM_GETCHECK, 0, 0))
		g_nStartupMode = ABK_MNGR_STARTUP_MIN;
	else if(SendDlgItemMessage(g_hMngrTabDlg, IDC_STARTUP_RADIO3, BM_GETCHECK, 0, 0))
		g_nStartupMode = ABK_MNGR_STARTUP_HIDE;

	g_fHideMin = (BOOL)SendDlgItemMessage(g_hMngrTabDlg, IDC_HIDE_MIN_CHECK,   BM_GETCHECK, 0, 0);
	g_fHideClose = (BOOL)SendDlgItemMessage(g_hMngrTabDlg, IDC_HIDE_CLOSE_CHECK, BM_GETCHECK, 0, 0);
	g_fOneClick = (BOOL)SendDlgItemMessage(g_hMngrTabDlg, IDC_ONE_CLICK_CHECK,  BM_GETCHECK, 0, 0);
	fStartup = (BOOL)SendDlgItemMessage(g_hMngrTabDlg, IDC_LOAD_AT_STARTUP,  BM_GETCHECK, 0, 0);
	g_fForceLogviewer = (BOOL)SendDlgItemMessage(g_hMngrTabDlg, IDC_FORCE_LOGVIEWER,  BM_GETCHECK, 0, 0);

	// Set or delete startup path.
	if(fStartup)
	{
		GetModuleFileName(g_hInst, szPath, MAX_PATH);
		cfRegSaveSTRVal(g_hKeyAS, REG_MNGR_AUTOSTART_VAL, szPath);
	}
	else
	{
		RegDeleteValue(g_hKeyAS, REG_MNGR_AUTOSTART_VAL);
	};

	nCurSel = (INT)SendDlgItemMessage(g_hMngrTabDlg, IDC_START_SWITCH, 
									  CB_GETCURSEL, 0, 0);
	g_nSwitchStart = (INT)SendDlgItemMessage(g_hMngrTabDlg, IDC_START_SWITCH, 
											 CB_GETITEMDATA, nCurSel, 0);

	nCurSel = (INT)SendDlgItemMessage(g_hMngrTabDlg, IDC_CLOSE_SWITCH, 
									  CB_GETCURSEL, 0, 0);
	g_nSwitchClose = (INT)SendDlgItemMessage(g_hMngrTabDlg, IDC_CLOSE_SWITCH, 
											 CB_GETITEMDATA, nCurSel, 0);

	// Password tab.
	lLen = cfGetDlgItemTextLength(g_hPwdTabDlg, IDC_FILE_NAME);
	cfdprintf("Text Len: %d", lLen);
	szBuf = new TCHAR[lLen+2];
	*szBuf = 0;
	GetDlgItemText(g_hPwdTabDlg, IDC_FILE_NAME, szBuf, lLen+1);	

	if(SendDlgItemMessage(g_hPwdTabDlg, IDC_PWD_RADIO1, BM_GETCHECK, 0, 0))
		g_nPasswdPlace = ABK_PASSWD_PLACE_SELF;
	else if(SendDlgItemMessage(g_hPwdTabDlg, IDC_PWD_RADIO2, BM_GETCHECK, 0, 0))
		g_nPasswdPlace = ABK_PASSWD_PLACE_ENV;
	else if(SendDlgItemMessage(g_hPwdTabDlg, IDC_PWD_RADIO3, BM_GETCHECK, 0, 0))
		g_nPasswdPlace = ABK_PASSWD_PLACE_REG;
	else if(SendDlgItemMessage(g_hPwdTabDlg, IDC_PWD_RADIO4, BM_GETCHECK, 0, 0))
		g_nPasswdPlace = ABK_PASSWD_PLACE_DEF_FILE;
	else if(SendDlgItemMessage(g_hPwdTabDlg, IDC_PWD_RADIO5, BM_GETCHECK, 0, 0))
		g_nPasswdPlace = ABK_PASSWD_PLACE_SELF_FILE;

	// Service tab.
	nCurSel = (INT)SendDlgItemMessage(g_hSvcTabDlg, IDC_SVC_START_SWITCH, CB_GETCURSEL, 0, 0);
	g_nSvcSwitchStart = (INT)SendDlgItemMessage(g_hSvcTabDlg, IDC_SVC_START_SWITCH, 
												CB_GETITEMDATA, nCurSel, 0);

	g_fUseStdPort = (BOOL)SendDlgItemMessage(g_hSvcTabDlg, IDC_USE_STD_PORT_CHECK,  BM_GETCHECK, 0, 0);
	g_fUseGWAddr = (BOOL)SendDlgItemMessage(g_hSvcTabDlg, IDC_USE_GATEWAY_CHECK,  BM_GETCHECK, 0, 0);
	g_nPortNumber = GetDlgItemInt(g_hSvcTabDlg, IDC_BKEYSRV_PORT, NULL, FALSE);
	g_fWriteLogFile = (BOOL)SendDlgItemMessage(g_hSvcTabDlg, IDC_WRITE_FILE,  BM_GETCHECK, 0, 0);
	g_fInformMngr = (BOOL)SendDlgItemMessage(g_hSvcTabDlg, IDC_INFORM_MNGR,  BM_GETCHECK, 0, 0);

	g_dwMaxEventsShow = GetDlgItemInt(g_hMngrTabDlg, IDC_LOG_CNT, NULL, FALSE);
	g_fPlaySound = (BOOL)SendDlgItemMessage(g_hMngrTabDlg, IDC_PLAYSOUND,  BM_GETCHECK, 0, 0);

	lLen = cfGetDlgItemTextLength(g_hSvcTabDlg, IDC_BKEYSRV_ADDR);
	delete[] g_szSrvAddress; // this is always allocated.
	g_szSrvAddress = new TCHAR[lLen+2];
	*g_szSrvAddress = 0;
	GetDlgItemText(g_hSvcTabDlg, IDC_BKEYSRV_ADDR, g_szSrvAddress, lLen+1);

	// Save variables.
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_STARTUP_MODE_VAL, (DWORD)g_nStartupMode));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_HIDE_MIN_VAL, (DWORD)g_fHideMin));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_HIDE_CLOSE_VAL, (DWORD)g_fHideClose));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_ONE_CLICK_VAL, (DWORD)g_fOneClick));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_FORCE_LOGVIEWER_VAL, (DWORD)g_fForceLogviewer));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_SWITCH_WHEN_CLOSE_VAL, (DWORD)g_nSwitchClose));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_SWITCH_WHEN_START_VAL, (DWORD)g_nSwitchStart));

	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_ASTERISKS_VAL, (DWORD)g_fAsterisks));

	cfCHECK_RET(cfRegSaveDWVal(g_hKeyLocMch, REG_PASSWD_PLACE_VAL, (DWORD)g_nPasswdPlace));
	cfCHECK_RET(cfRegSaveSTRVal(g_hKeyLocMch, REG_PASSWD_FILE_VAL, szBuf));

	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_LOGVIEWER_MAX_ENTRY, g_dwMaxEventsShow));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyCurUsr, REG_LOGVIEWER_PLAYSOUND, g_fPlaySound));

	cfCHECK_RET(cfRegSaveDWVal(g_hKeyLocMch, REG_SWITCH_WHEN_SVC_START_VAL, (DWORD)g_nSvcSwitchStart));

	cfCHECK_RET(cfRegSaveDWVal(g_hKeyLocMch, REG_BKEY_SRV_PORT_USE_STD, (DWORD)g_fUseStdPort));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyLocMch, REG_BKEY_SRV_USE_DEF_GW_ADDR, (DWORD)g_fUseGWAddr));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyLocMch, REG_BKEY_SRV_PORT, (DWORD)g_nPortNumber));
	cfCHECK_RET(cfRegSaveSTRVal(g_hKeyLocMch, REG_BKEY_SRV_ADDR, g_szSrvAddress));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyLocMch, REG_BKEY_SRV_WRITE_LOG, (DWORD)g_fWriteLogFile));
	cfCHECK_RET(cfRegSaveDWVal(g_hKeyLocMch, REG_BKEY_SRV_INFORM_MNGR, (DWORD)g_fInformMngr));

	delete[] szBuf;

	// Send update configuration to the service.
	if(g_fSvcConfChanged)
	{
		cfdprintf("Service config changed, tell service about...");

		RpcTryExcept
		{
			SvcUpdateConfiguration();
		}
		RpcExcept(1)
		{
			//
		}
		RpcEndExcept;
	};

	return(TRUE);
};

VOID DoExitManager(VOID)
{
	g_fExit = TRUE;

	RpcTryExcept
	{
		// If user wants to change access mode when manager starts, 
		// it is a proper place do do it.
		if(g_nSwitchClose > ABK_ACCESS_LEAVE)
			SetAccessMode(g_nSwitchClose);

		SvcDisconnectManager();
	}
	RpcExcept(1)
	{
		DestroyWindow(g_hMainDlg);
	}
	RpcEndExcept;
};

// This routine checks for new fersion every one hour.
// It is called by the system. 
//VOID CALLBACK CheckForNewVersionProc(HWND hwnd_, UINT nMsg_, UINT_PTR idEvent_, DWORD dwTime_)
unsigned int __stdcall CheckForNewVersionProc(void*)
{
	HINTERNET	hInet;
	HINTERNET	hUrl;
	CHAR		cBuff[BKP_UPDATE_BUFF_SIZE];
	CHAR*		ptr;
	DWORD		dwRead;

	cfdprintf(_T("CheckForNewVersionProc"));

	KillTimer(g_hMainDlg, BKP_UPDATE_TIMER_ID);

	hInet = InternetOpen(_T("BotikKey Pro"), INTERNET_OPEN_TYPE_PRECONFIG, 
						 NULL, NULL, 0);

	if(hInet == NULL)
	{
		cfdprintf(_T("CheckForNewVersionProc: Failed to open internet handle."));
		goto finish;
	};

	hUrl = InternetOpenUrl(hInet, BKP_UPDATE_URL, NULL, 0, 
							  INTERNET_FLAG_NO_CACHE_WRITE 
							| INTERNET_FLAG_PRAGMA_NOCACHE 
							| INTERNET_FLAG_RELOAD, 0);

	if(hUrl == NULL)
	{
		cfdprintf(_T("CheckForNewVersionProc: Failed to open URL."));
		goto cleanup1;
	};

	ptr = cBuff;

	ZeroMemory(cBuff, sizeof(cBuff));

	InternetReadFile(hUrl, ptr, BKP_UPDATE_BUFF_SIZE, &dwRead);

	cfdprintf(_T("CheckForNewVersionProc: Got from web-server: %s"), cBuff);

	// Validate data.
	if(strstr(cBuff, BKP_UPDATE_PREFIX) != ptr)
	{
		cfdprintf(_T("CheckForNewVersionProc: Bad data."));
	}
	else
	{
		TCHAR	szPath[MAX_PATH+1];
		DWORD	dwSize;
		DWORD	dwHandle;

		cfdprintf(_T("CheckForNewVersionProc: Good data."));

		GetModuleFileName(g_hInst, szPath, MAX_PATH);

		// Recognize version.
		dwSize = GetFileVersionInfoSize(szPath, &dwHandle);

		if(dwSize == 0)
		{
			cfdprintf(_T("CheckForNewVersionProc: GetVersionInfoSize failed."));
		}
		else
		{
			LPBYTE	lpBuff;

			lpBuff = new BYTE[dwSize];

			if(!GetFileVersionInfo(szPath, dwHandle, dwSize, lpBuff))
			{
				cfdprintf(_T("CheckForNewVersionProc: GetFileVersionInfo failed."));
			}
			else
			{
				VS_FIXEDFILEINFO* lpffi;
				UINT nLen;

				if(VerQueryValue(lpBuff, _T("\\"), (LPVOID*)&lpffi, &nLen))
				{
					INT nNewBuild;

					cfdprintf(_T("CheckForNewVersionProc: Executable version: %d.%d.%d.%d"), HIWORD(lpffi->dwProductVersionMS), LOWORD(lpffi->dwProductVersionMS), HIWORD(lpffi->dwProductVersionLS), LOWORD(lpffi->dwProductVersionLS));

					ptr = strrchr(cBuff, '.');
					ptr++;

					if(ptr)
					{
						nNewBuild = atoi(ptr);
						cfdprintf(_T("CheckForNewVersionProc: Build number of current version: %d"), nNewBuild);

						if(   nNewBuild != 0 
						   && nNewBuild > LOWORD(lpffi->dwProductVersionLS) 
						   && nNewBuild > g_nLastNewVer)
						{
							LPTSTR szMsg;

							cfdprintf("CheckForNewVersionProc: There is a new version.");
							
							// XXX
							szMsg = _T("      BotikKey Pro.");

							if(g_fIs2K && g_fBlnEnbl)
							{
								ShowBalloonNotifyIcon(g_hMainDlg, ABK_MNGR_NOTIFY_ICON_ID, NIIF_INFO, 
													  ABK_MANAGER_NAME, szMsg);
							}
							else
							{
								INT nRet;

								SetForegroundWindow(g_hMainDlg);

								nRet = MessageBox(g_hMainDlg, szMsg, ABK_MANAGER_NAME, MB_OK/*MB_YESNO|*/|MB_ICONASTERISK);

								if(nRet == IDNO)
									g_nLastNewVer = nNewBuild;
							};
						};
					};
				}; // if(VerQueryValue(lpBuff, _T("\\"), (LPVOID*)&lpffi, &nLen))
			}; // else of if(!GetFileVersionInfo(szPath, dwHandle, dwSize, lpBuff))

			delete[] lpBuff;
		}; // else of if(dwSize == 0)
	}; // else of if(strstr(cBuff, BKP_UPDATE_PREFIX) != ptr)

	InternetCloseHandle(hUrl);

cleanup1:
	InternetCloseHandle(hInet);

finish:
	SetTimer(g_hMainDlg, BKP_UPDATE_TIMER_ID, BKP_UPDATE_TIMER_ELAPSE, (TIMERPROC)NULL);

	cfdprintf(_T("!CheckForNewVersionProc"));

	return(0);
};


