/*  BotikKey Pro - Botik Key Client for Windows 9x/ME/NT/2000/XP/2003
 *  Copyright (C) 2004 Alexander V. Melnikoff
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */

#include "stdafx.h"
#include <process.h>
#include <time.h>
#include "service.h"
#include "interface_h.h"
#include "rpc_procs.h"
#include "eventlog.h"
#include "bkeyclient.h"

// XXX Temporarily added. 
void koi8r_to_win(unsigned char* szBuf_);

#define ABK_SVC_BKEY_SLEEP_TIME			5*1000 // Sleep for a 5 seconds before trying to reconnect to server.

// Commands from server
#define ABK_SVC_BKEY_CMD_BOTIKKEY		0x01
#define ABK_SVC_BKEY_CMD_MESSAGE		0x02
#define ABK_SVC_BKEY_CMD_CHALLENGE		0x04
#define ABK_SVC_BKEY_CMD_GOOD			0x08
#define ABK_SVC_BKEY_CMD_SORRY			0x10

HANDLE		g_hClientThrd;
HANDLE		g_hNetEvt;
HANDLE		g_hChangeModeEvt;
HANDLE		g_hShutdownEvt;

DWORD		g_dwStatus;
DWORD		g_dwTimeOut;

LPSTR		szExpected;			// Expected commands from server.
SOCKET		hSock;				// Socket
SOCKADDR_IN	saiAddrSrv;

UINT  WINAPI __BKClientThrd(void*);
BOOL  BKSocketInit();
LPSTR BKSocketRead(LPDWORD pdwSize_);
BOOL  BKProcessData(LPSTR szBuf_);

BOOL BKChallengeCmd(LPSTR szTxt_);

BOOL StartBotikKeyClient(VOID)
{
	unsigned int thrdid;

	g_hClientThrd = (HANDLE)_beginthreadex(NULL, 0, __BKClientThrd, NULL, 0, &thrdid);

	cfdprintf("StartBotikKeyClient: ret: %d", g_hClientThrd);

	if(g_hClientThrd == (HANDLE)-1)
		return(FALSE);

	return(TRUE);
};

unsigned int __stdcall __BKClientThrd(void*)
{
	LONG	lRet;
	HANDLE	hHandles[3];
	DWORD	dwEvent;
	BOOL	fLoop;
	DWORD	dwEvtNum;

	fLoop = TRUE;
	hSock = NULL;

	// Preparing to wait.
	hHandles[0] = g_hChangeModeEvt;
	hHandles[1] = g_hShutdownEvt;
	hHandles[2] = g_hNetEvt;

	md5_init();

	g_dwStatus = ABK_SVC_CONNECTING;

	g_dwTimeOut = ABK_BKEY_SRV_CON_TIMEOUT;

	while(fLoop)
	{
		// Init connection.
		if(hSock == NULL)
		{
			cfdprintf(_T("__BKClientThrd: hSock == NULL, try to init socket."));

			// If key-server address (!g_szSrvAddr[0]) is not set, 
			// try to get it again. 
			if(!g_szSrvAddr[0])
			{
				cfdprintf(_T("__BKClientThrd: Bad key-server address, try to get it."));

				GetServerAddress();
			};

			// Check if we have a key-server address, we can init connection.
			// If we have not a key-server address, we should sleep for a while
			// and then try to open connection again.
			if(g_szSrvAddr[0])
			{
				cfdprintf(_T("__BKClientThrd: Good key-server address, try to init socket."));

				if(!BKSocketInit())
					break;

				szExpected = "BOTIKKEY";

				g_dwTimeOut = ABK_BKEY_SRV_CON_TIMEOUT;
				dwEvtNum = 3;
			}
			else
			{
				cfdprintf(_T("__BKClientThrd: Get Key-server address failed, so sleep for a while."));

				dwEvtNum = 2;
				g_dwTimeOut = 1 * 60 * 1000; // Wait for 1 minute

				EventLogAddEntryF(BKP_EVENT_TYPE_ERROR, 
						_T("    Key-.   %d ."), 
						g_dwTimeOut/1000);
			};
		}; // if(hSock == NULL)

		cfdprintf(_T("__BKClientThrd: Begin WaitForMultipleObjects (num: %d; tout: %dms)..."), dwEvtNum, g_dwTimeOut);

		dwEvent = WaitForMultipleObjects(dwEvtNum, hHandles, FALSE, g_dwTimeOut);
		if(dwEvent == WAIT_FAILED)
		{
			// XXX Error...
			cfdprintf(_T("__BKClientThrd: WaitForMultipleObjects failed."));

			return(0);
		}
		else if(dwEvent == WAIT_TIMEOUT)
		{
			cfdprintf(_T("__BKClientThrd: WaitForMultipleObjects timed out."));

			// If timed out, close connection, and try to reconnect again.
			closesocket(hSock);
			g_dwStatus = ABK_SVC_CONNECTING;
			g_dwAccessMode |= ABK_ACCESS_CHANGING;
			SetEvent(g_hDispMsg01Evt);
			SetEvent(g_hDispMsg03Evt);
			hSock = NULL;

			EventLogAddEntry(BKP_EVENT_TYPE_INFORMATION, 
							 _T("    Key- ,       "));
		}
		else if(dwEvent == WAIT_OBJECT_0) // We got message from manager.
		{
			// To set new access mode we should close current connection
			// and open it again.
			cfdprintf(_T("__BKClientThrd: WAIT_OBJECT_0: Set new access mode event fired."));

			closesocket(hSock);
			g_dwStatus = ABK_SVC_CONNECTING;
			g_dwAccessMode |= ABK_ACCESS_CHANGING;
			SetEvent(g_hDispMsg01Evt);
			SetEvent(g_hDispMsg03Evt);
			hSock = NULL;
		}
		else if(dwEvent == (WAIT_OBJECT_0 + 1))
		{
			// Shutdown client thread.
			cfdprintf(_T("__BKClientThrd: WAIT_OBJECT_1: Shutdown event fired."));

			fLoop = FALSE;
		}
		else // We got event from WinSock.
		{
			WSANETWORKEVENTS wsane;

			cfdprintf(_T("__BKClientThrd: WAIT_OBJECT_2: Network event fired."));

			// Get to know which event we got.
			lRet = WSAEnumNetworkEvents(hSock, g_hNetEvt, &wsane);
			if(lRet == SOCKET_ERROR)
				fLoop = FALSE;

			if(wsane.lNetworkEvents & FD_READ) 
			{
				// We got a message from server. Read it and process.
				DWORD dwSize;
				LPSTR szBuf;

				cfdprintf(_T("__BKClientThrd: FD_READ event from socket."));

				szBuf = BKSocketRead(&dwSize);
				szBuf[dwSize] = 0;
				
				if(!BKProcessData(szBuf))
				{
					// Data processing failed. Close connection and reconnect again.
					closesocket(hSock);
					g_dwStatus = ABK_SVC_CONNECTING;
					g_dwAccessMode |= ABK_ACCESS_CHANGING;
					SetEvent(g_hDispMsg01Evt);
					SetEvent(g_hDispMsg03Evt);
					hSock = NULL;
				};

				delete[] szBuf;
			};

			if(wsane.lNetworkEvents & FD_CLOSE)
			{
				// The server closed socket. Try to connect again.
				cfdprintf(_T("__BKClientThrd: FD_CLOSE event from socket."));

				closesocket(hSock);
				g_dwReason = 0;
				g_dwStatus = ABK_SVC_CONNECTING;
				g_dwAccessMode |= ABK_ACCESS_CHANGING;
				SetEvent(g_hDispMsg01Evt);
				SetEvent(g_hDispMsg03Evt);
				hSock = NULL;
			};

			if(wsane.lNetworkEvents & FD_CONNECT) 
			{										   
				// Connection event, we are connected or refused. 
				// If connection refused, try to reconnect. If connected, do nothing.
				cfdprintf(_T("__BKClientThrd: FD_CONNECT event from socket. Err:%d"), wsane.iErrorCode[FD_CONNECT_BIT]);

				g_dwReason = 0;

				if(wsane.iErrorCode[FD_CONNECT_BIT])
				{
					closesocket(hSock);
					hSock = NULL;

					g_dwReason = ABK_SVC_CONNECT_FAILED;

					// XXX Localization.
					EventLogAddEntryF(BKP_EVENT_TYPE_ERROR, 
						_T("    Key- %s:%d.   %d ."), 
						g_szSrvAddr, g_nPortNum, (ABK_SVC_BKEY_SLEEP_TIME/1000));

					Sleep(ABK_SVC_BKEY_SLEEP_TIME);
				};

				SetEvent(g_hDispMsg01Evt);
				SetEvent(g_hDispMsg03Evt);
			};
		}; // if(dwEvent == WAIT_FAILED)
	}; // while(fLoop)

	md5_fin();

	closesocket(hSock);

	return(0);
};

#define ABK_BKSR_CHUNK_SIZE				512
LPSTR BKSocketRead(LPDWORD pdwSize_)
{
	DWORD dwSize = 0;
	LPSTR outptr = NULL;
	BOOL fLoop;
	INT nRcv;

	fLoop = TRUE;

	while(fLoop)
	{
		CHAR buf[ABK_BKSR_CHUNK_SIZE+1];

		nRcv = recv(hSock, buf, ABK_BKSR_CHUNK_SIZE, 0);

		if(nRcv == SOCKET_ERROR || nRcv == 0) // We got an error or no data (nRcv == -1 || nRcv == 0).
		{
			cfdprintf("BKSocketRead: nRcv: %d", nRcv);
			fLoop = FALSE;
		}
		else
		{
			if(outptr == NULL)
			{
				// Alloc a memory.
				outptr = new CHAR[nRcv+1];
			}
			else
			{
				// Realloc memory.
				LPSTR ptr = new CHAR[dwSize + nRcv + 1];
				
				memcpy(ptr, outptr, dwSize); // Copy data to the new buffer
				delete[] outptr; // Free old buffer.
				outptr = ptr;
			};

			// Copy received data into the output buffer.
			memcpy(outptr+dwSize, buf, nRcv);
			dwSize += nRcv;
		};
	}; // while(fLoop)

	*pdwSize_ = dwSize;

	return(outptr);
};

BOOL BKSocketInit()
{
	LONG	lRet;
	
	// Create socket.
	hSock = WSASocket(AF_INET, SOCK_STREAM, 0, NULL, 0, 0);
	if(hSock == INVALID_SOCKET)
	{
		// XXX Error...
		cfdprintf("BKSocketInit: WSASocket failed.");
		return(FALSE);
	};

	ZeroMemory((PVOID)&saiAddrSrv, sizeof(SOCKADDR_IN));
	saiAddrSrv.sin_family = AF_INET;
	saiAddrSrv.sin_addr.s_addr = inet_addr(g_szSrvAddr);
	saiAddrSrv.sin_port = htons(g_nPortNum);

	// Bind socket and network event object together.
	lRet = WSAEventSelect(hSock, g_hNetEvt, FD_READ|FD_CLOSE|FD_CONNECT);
	if(lRet == SOCKET_ERROR)
	{
		// XXX Error...
		cfdprintf("BKSocketInit: WSAEventSelect failed.");
		return(FALSE);
	};

	lRet = WSAConnect(hSock, (SOCKADDR*)&saiAddrSrv, sizeof(SOCKADDR_IN), NULL, NULL, NULL, NULL);
	INT err = WSAGetLastError();

	cfdprintf("BKSocketInit: lRet: %d; err: %d", lRet, err);

	if(lRet == SOCKET_ERROR && err != WSAEWOULDBLOCK)
	{
		// XXX Error...
		cfdprintf("BKSocketInit: WSAConnect failed.");
		return(FALSE);
	};

	return(TRUE);
};

BOOL BKProcessData(LPSTR szBuf_)
{
	LPSTR line = szBuf_;

	line = strtok(szBuf_, "\n\r");
	while(line)
	{
		cfdprintf("BKProcessData: Line {%s}", line);

		LPSTR str, cmd = line;

		// line += strcspn(line, "\t ");
		while(*line >= 'A' && *line <= 'Z') line++; // Skip command characters.

		*line = 0; line++;

		while(*line == '\t' || *line == ' ') line++; // Skip white space sharacters.

		str = line;

		cfdprintf("BKProcessData: CMD: %s; STR: %s", cmd, str);

		if(*cmd == 0 || strstr(szExpected, cmd) == NULL)
		{
			if(strlen(cmd) > 0)
			{
				cfdprintf("BKProcessData: protocol error: expected %s, but got %s", szExpected, cmd);

				EventLogAddEntryF(BKP_EVENT_TYPE_ERROR, 
					_T("  --   Key- %s, : %s"), 
					szExpected, cmd);
			};

			return(FALSE);
		}
		else
		{
			if(!strcmp(cmd, "BOTIKKEY"))
			{
				LPSTR ptr;

				cfdprintf("BKProcessData: Connected to the KeyServer %s %s", cmd, str);

				EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, 
								 _T("  Key-: %s %s"), cmd, str);

				if(ptr = strstr(str, "TIMEOUT="))
				{
					DWORD dwVal;

					ptr += strlen("TIMEOUT=");
					dwVal = atoi(ptr);

					cfdprintf("BKProcessData: Timeout from server: %d", dwVal);

					EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, 
									  _T("Key-    : %d ."), dwVal);

					if(dwVal > 0)
						g_dwTimeOut = dwVal*1000;
				};

				szExpected = "MESSAGE CHALLENGE";

				cfdprintf(_T("BKProcessData: Sending message about connected status to the mngr."));

				g_dwStatus = ABK_SVC_CONNECTED;
				g_dwReason = 0;
				SetEvent(g_hDispMsg03Evt);
			}
			else if(!strcmp(cmd, "MESSAGE"))
			{
				koi8r_to_win((UCHAR*)str);

				cfdprintf("BKProcessData: Got a message from ST Botik: %s", str);

				EventLogAddEntryF(BKP_EVENT_TYPE_MESSAGE, 
								 _T("  : %s"), str);

				// We do not need to change szExpected.
			}
			else if(!strcmp(cmd, "CHALLENGE"))
			{
				cfdprintf("BKProcessData: Got a CHALLENGE message..."/*, str*/);

				EventLogAddEntry(BKP_EVENT_TYPE_INFORMATION, 
								 _T("   (challenge)..."));

				if(!BKChallengeCmd(str))
				{
					return(FALSE);
				};
				szExpected = "GOOD SORRY MESSAGE";
			}
			else if(!strcmp(cmd, "GOOD"))
			{
				DWORD dwModeBySrv;
				CHAR  szModeBySrv[6];
				LPSTR ptr;

				cfdprintf("BKProcessData: Got a GOOD message: %s", str);

				//cfBREAKPT

				// Extract access mode from server string.
				ptr = str;
				while(*ptr >= 'A' && *ptr <= 'Z') ptr++; // Skip mode characters.
				strncpy(szModeBySrv, str, ptr-str);
				szModeBySrv[ptr-str] = 0; // Set term zero.

				cfdprintf("BKProcessData: szModeBySrv: [%s]", szModeBySrv);

				dwModeBySrv = StringToAccessMode(szModeBySrv);

				cfdprintf("BKProcessData: dwModeBySrv: %d; g_dwReqAccessMode: %d", dwModeBySrv, g_dwReqAccessMode);

				g_dwAccessMode = dwModeBySrv;

				if(dwModeBySrv != g_dwReqAccessMode)
				{
					cfdprintf("BKProcessData: We got GOOD but requested mode(%s) not eq got mode(%s)", 
						AccessModeToString(g_dwReqAccessMode), str);

					EventLogAddEntryF(BKP_EVENT_TYPE_OK, 
						_T(" %s,   ,    %s"), 
						AccessModeToString(g_dwReqAccessMode), str);
				}
				else
				{
					cfdprintf("BKProcessData: Our responce is accepted (%s %s).", cmd, str);

					EventLogAddEntryF(BKP_EVENT_TYPE_OK, 
						_T(" %s,   , Key- : %s %s"), 
						AccessModeToString(g_dwReqAccessMode), cmd, str);
				};

				// Send message about current access mode to the manager if it connected.
				if(g_fIsMsgDisp)
				{
					SetEvent(g_hDispMsg01Evt);
				};

				szExpected = "MESSAGE CHALLENGE";
			}
			else if(!strcmp(cmd, "SORRY"))
			{
				cfdprintf("BKProcessData: Got a SORRY message: %s %s", cmd, str);
				
				EventLogAddEntryF(BKP_EVENT_TYPE_ERROR, 
						_T(" %s,   , Key- : %s"), 
						AccessModeToString(g_dwReqAccessMode), cmd);

				g_dwAccessMode = ABK_ACCESS_NONE;

				// Send message about current access mode to the manager if it connected.
				if(g_fIsMsgDisp)
				{
					SetEvent(g_hDispMsg01Evt);
				};

				szExpected = "MESSAGE CHALLENGE";
			};
		};

		line = strtok(NULL, "\n\r");
	};

	return(TRUE);
};

BOOL BKChallengeCmd(LPSTR szTxt_)
{
	CHAR	szMD5Rnd[32+1];
	CHAR	szMD5Tok[32+1];
	LPSTR	szTok;
	LPSTR	szResp;
	LONG	lLen;
	LPSTR	szAccess;
	INT		nRet;

	UCHAR*	szUuid;
	UUID	uuid;

	srand((unsigned)time(NULL));

	// Get a random hash.
	UuidCreate(&uuid);
	UuidToStringA(&uuid, &szUuid);

	md5_md5((LPBYTE)szUuid, (ULONG)strlen((CHAR*)szUuid), szMD5Rnd);

	//cfdprintf(_T("UUID: %s; MD5: %s"), szUuid, szMD5Rnd);

	RpcStringFree(&szUuid);

	// Desired access as string.
	if(g_dwReqAccessMode == ABK_ACCESS_UNSPEC) // XXX For some case...
		g_dwReqAccessMode = ABK_ACCESS_NONE;

	szAccess = AccessModeToString(g_dwReqAccessMode);

	cfdprintf("BKChallengeCmd: Access: %s; ReqAccess: %d", szAccess, g_dwReqAccessMode);

	// Prepare hash token.
	lLen = (LONG)(strlen(szTxt_)+strlen(szMD5Rnd)+strlen(szAccess)+strlen(g_szBotikKey));
	szTok = new CHAR[lLen+1];
	sprintf(szTok, "%s%s%s%s", szTxt_, szMD5Rnd, szAccess, g_szBotikKey);
	md5_md5((LPBYTE)szTok, (ULONG)strlen(szTok), szMD5Tok);

	//cfdprintf("Token: %s -> %s", szTok, szMD5Tok);

	delete[] szTok;

	// Prepare response command.
	lLen = (LONG)(strlen(szMD5Tok)+strlen(szMD5Rnd)+strlen(szAccess));
	szResp = new CHAR[lLen + 8 + 5]; // 9 is a length of RESPONSE command and 4 is spaces and term zero.
	sprintf(szResp, "RESPONSE %s %s %s\n", szMD5Tok, szMD5Rnd, szAccess);

	//cfdprintf("Response: %s", szResp);

	EventLogAddEntryF(BKP_EVENT_TYPE_INFORMATION, 
		_T("   (response),  : %s"), 
		szAccess);

	nRet = send(hSock, szResp, (INT)strlen(szResp), 0);

	delete[] szResp;

	if(nRet == SOCKET_ERROR)
	{
		INT nCode = WSAGetLastError();

		EventLogAddEntryF(BKP_EVENT_TYPE_ERROR, 
			_T("   .  : %d"), 
			nCode);

		return(FALSE);
	};

	return(TRUE);
};
